"""
Validadores generales.

Por el momento incluye validadores a nivel de imágenes y archivos.
Ver más info en:

https://docs.djangoproject.com/en/2.0/ref/validators/

List of MIME types: https://www.sitepoint.com/mime-types-complete-list/

Cómo usarlo:

Se recomienda insertar las restricciones en el archivo ``settings.py``:

    # Para los validadores de fichero
    MAX_UPLOAD_SIZE = 1048576   # 1Mb
    ALLOWED_IMAGE_MIME_TYPES = ['image/jpeg', 'image/png', 'image/gif']
    ALLOWED_VIDEO_MIME_TYPES = ['video/mpeg', 'video/quicktime', 'video/mp4']

Supongamos que tenemos un modelo de elemento que contiene un atributo ``media``
que puede contener un vídeo o una imagen. Lo definimos entonces como de tipo
``FileField``:

.. code-block:: python

    class Element(models.Model):

        def get_upload_path(instance, filename):
            ...

        ...
        media = models.FileField(
            upload_to=get_upload_path, null=True, blank=True,
            validators=[ContentTypeValidator(
                accepted_types=settings.ELEMENTS_SETTINGS['ALLOWED_IMAGE_MIME_TYPES'] +
                               settings.ALLOWED_VIDEO_MIME_TYPES),
                MaxFileSizeValidator(max_size=settings.ELEMENTS_SETTINGS['MAX_UPLOAD_SIZE'])])
        ...

En el ejemplo solamente se aceptarían archivos de tipos MIME contenidos en
``ALLOWED_IMAGE_MIME_TYPES`` y ``ALLOWED_VIDEO_MIME_TYPES``.

Así mismo, solamente se aceptan imágenes de máximo 1Mb de tamaño. Date cuenta
que el tamaño está especificado en bytes: 1024 x 1024 = 1048576.
"""

from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.utils.deconstruct import deconstructible


@deconstructible
class ContentTypeValidator:

    """
    Valida que el `content_type` de un archivo esté contenido en la lista
    `accepted_types`.
    Debe ser serializable para que funcione con las migraciones:
    https://docs.djangoproject.com/en/2.0/ref/validators/
    """

    message = 'Content type not accepted'
    code = 'content_type'

    def __init__(self, accepted_types):
        self.accepted_types = accepted_types

    def __call__(self, file):
        # Solamente validamos si el archivo es nuevo
        # y no si el archivo ya estaba en el campo
        # de una operación anterior, en cuyo caso
        # ya estaría validado
        if type(file.file) == InMemoryUploadedFile:
            content_type = file.file.content_type
            if content_type not in self.accepted_types:
                raise ValidationError(message=self.message, code=self.code)


@deconstructible
class MaxFileSizeValidator:

    """
    Valida que un archivo sea menor que el tamaño máximo permitido.
    El tamaño se especifica en bytes.
    Debe ser serializable para que funcione con las migraciones:
    https://docs.djangoproject.com/en/2.0/ref/validators/
    """

    message = 'File is too large'
    code = 'max_file_size'

    def __init__(self, max_size):
        self.max_size = max_size

    def __call__(self, file):
        if self.max_size < file.size:
            raise ValidationError(message=self.message, code=self.code)



from rest_framework.pagination import PageNumberPagination
from django.conf import settings


class DefaultPagination(PageNumberPagination):
    """
    Clase Paginator específica para fijar la paginación en una query.
    El funcionamiento será mediante el paso de parámetros se fijará el número
    máximo de elementos con los que paginaremos.
    Con esta estrategia podremos solicitar en la propia query el número máximo
    de elementos por página.

    Para usar esta paginación es necesario especificar las settigs de
    paginación en el archivo de settings del proyecto:

    ```
    # Settings para la paginación del proyecto
    PAGINATION_SETTINGS = {'DEFAULT_LIMIT': 50, 'MAX_LIMIT_DEFAULT': 20000}
    ```
    """

    page_size = settings.PAGINATION_SETTINGS["DEFAULT_LIMIT"]
    page_size_query_param = "page_size"
    max_page_size = settings.PAGINATION_SETTINGS["MAX_LIMIT_DEFAULT"]



class SoftDeletionModel(models.Model):
    """
    Modelo abstracto para la gestión de la eliminación
    *ficticia* de entidades.

    Cualquier clase que herede de ``SoftDeletionModel`` permite realizar un
    borrado `virtual`, es decir, los objetos no se eliminan de la base de datos
    sino que se marcan como eliminados en una operación estándar de borrado.

    Por defecto, cualquier consulta realizada sobre la colección ``objects``
    devolverá solamente aquellos objetos que no están marcados como eliminados.

    Para permitir el acceso a lo elementos marcados como eliminados,
    proporciona la colección ``all_objects`` que accede a **TODOS** los
    elementos. En realidad, no es más que un alias de la colección ``objects``
    original.

    Asimismo, proporciona el método ``hard_delete`` para la eliminación
    real (a nivel de base de datos) del objeto.

    Por ejemplo, supongamos que tenemos definido el model ``MyModel``, definido
    como cualquier otro modelo de Django, pero que hereda de
    ``SoftDeletionModel``:

    .. code-block::python

        from common.models import SoftDeletionModel


        class MyModel(SoftDeletionModel):
            ...


    - ``MyModel.objects.get(pk=123).delete()`` no eliminará el objeto de la
      base de datos, sino que actualizará el atributo ``deleted_at`` con la
      fecha y hora actual. Esto lo marca como eliminado.

    - ``MyModel.objects.all()`` devolverá todos los objetos que no tengan valor
      en el atributo ``deleted_at``, es decir, todos los que no están marcados
      como eliminados.

    - Si llamamos a ``MyModel.objects.get(pk=123)`` después de la llamada a
      ``delete`` obtendremos una excepción ``ObjectDoesNotExist``. En cambio,
      si llamamos a ``MyModel.all_objects.get(pk=123)`` sí que obtendremos
      el objeto marcado como eliminado. Evidentemente esto nos permite hacer un
      *restaurar* o *undelete* simplemente eliminando el valor del atributo
      ``deleted_at``, es decir, estableciendo su valor a ``None``.
    """

    deleted_at = models.DateTimeField(blank=True, null=True)

    objects = SoftDeletionManager()
    all_objects = SoftDeletionManager(alive_only=False)

    class Meta:
        abstract = True

    def delete(self, using=None, keep_parents=False):
        """
        Marca un objeto como eliminado, es decir, realiza la operación de
        borrado `ficticio`.
        """
        self.deleted_at = timezone.now()
        self.save()

    def hard_delete(self, using=None, keep_parents=False):
        """
        Realiza la operación de borrado definitivo (no `ficticio`).
        """
        super(SoftDeletionModel, self).delete(using=using, keep_parents=keep_parents)




"""
Define algunos permisos estándar para utilizar a lo largo de toda la
aplicación.
"""
import logging

from rest_framework.permissions import BasePermission


logger = logging.getLogger(__name__)


class IsOwnerAdminOrReadOnly(BasePermission):
    """
    Permite la lectura a través del método ``GET`` para
    cualquier usuario.

    Cualquier otro tipo de operación estará limitada a que el usuario que
    realiza la petición sea usuario administrador o que coincida con el valor
    del atributo ``user`` del objeto.
    """

    def has_object_permission(self, request, view, obj):
        if request.method == "GET":
            return True
        return obj.user == request.user or request.user.is_superuser


class IsOwnerOrAdmin(BasePermission):
    """
    Cualquier operación de acceso estará limitada a que el usuario que
    realiza la petición sea usuario administrador o que coincida con el valor
    del atributo ``user`` del objeto.
    """

    def has_object_permission(self, request, view, obj):
        logger.debug("Current user: %s", request.user)
        logger.debug(
            "Is current user authenticated?: %s", request.user.is_authenticated
        )
        logger.debug("Is current user an admin user?: %s", request.user.is_superuser)
        return request.user.is_authenticated and (
            obj.user == request.user or request.user.is_superuser
        )



"""
Querysets personalizados y comunes para usar en cualquier app.
"""

from django.utils import timezone
from django.contrib.gis.db.models import QuerySet


class SoftDeletionQuerySet(QuerySet):
    """
    QuerySet específico para el borrado `ficticio`.
    Proporciona soporte para realizar el borrado suave o ficticio y
    el borrado definitivo, así como métodos para obtener los objetos
    borrados y no borrados.
    """

    def delete(self):
        """
        Sobreescribe el método ``delete`` por defecto de un objeto actualizando
        la fecha de borrado para realizar el borrado `ficticio`.
        """
        return super(SoftDeletionQuerySet, self).update(deleted_at=timezone.now())

    def hard_delete(self):
        """
        Elimina definitivamente un objeto. Corresponde con el comportamiento
        por defecto del método ``delete``.
        """
        return super(SoftDeletionQuerySet, self).delete()

    def alive(self):
        """
        Helper que filtra los elementos que no están marcados como borrados.
        Posiblemente no resulte de mucha utilidad, pero se usa internamente
        por el sistema de borrado.
        """
        return self.filter(deleted_at=None)

    def dead(self):
        """
        Helper que filtra los elementos que están marcados como borrados.
        Posiblemente no resulte de mucha utilidad, pero se usa internamente
        por el sistema de borrado.
        """
        return self.exclude(deleted_at=None)



"""
Managers personalizados comunes para usar a lo largo de todas las apps.
"""

from django.contrib.gis.db import models

from .querysets import SoftDeletionQuerySet


class SoftDeletionManager(models.Manager):
    """
    Manager para gestionar el borrado `ficticio`.
    """

    def __init__(self, *args, **kwargs):
        self.alive_only = kwargs.pop("alive_only", True)
        super(SoftDeletionManager, self).__init__(*args, **kwargs)

    def get_queryset(self):
        """
        Devuelve un QuerySet con solamente los objetos marcados como borrados
        o todos en función del valor del atributo ``alive_only``.
        """
        if self.alive_only:
            return SoftDeletionQuerySet(self.model).filter(deleted_at=None)
        return SoftDeletionQuerySet(self.model)

    def hard_delete(self):
        """
        Realiza una eliminación `masiva` y definitiva (no `ficticia`) de
        los elementos del queryset.
        """
        return self.get_queryset().hard_delete()
